#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
import datetime
import operator
from functools import reduce
from typing import List, Sequence

from hypothesis import strategies as st
from hypothesis.strategies import SearchStrategy
from xotl.tools.infinity import Infinity
from xotl.tools.objects import DelegatedAttribute

from ...domains import EquivalenceSet, IntervalSet, TOrd
from ...ranges import Excluded, Included, Range

# Notice that floats have a maximum.  At the time of writing::
#    >>> l=9007199254740992.0
#    >>> l == l + 1   # Really?!!
#    True
#
# Since we create ranges like Range(lower, lower + delta), let's put a cap on
# the following generators so that generated ranges have a sensible and
# predictable behavior.
numbers = st.floats(
    allow_infinity=False, allow_nan=False, min_value=-10000.0, max_value=10000.0
)
nonnegative_numbers = st.floats(
    allow_infinity=False, allow_nan=False, min_value=0, max_value=10000.0
)
negative_numbers = st.floats(
    allow_infinity=False, allow_nan=False, min_value=-10000.0, max_value=-1
)
positive_numbers = st.floats(
    allow_infinity=False, allow_nan=False, min_value=1, max_value=10000.0
)


equivalence_sets: SearchStrategy[EquivalenceSet[int]] = st.frozensets(
    st.integers()
).map(EquivalenceSet)
non_empty_equivalence_sets: SearchStrategy[EquivalenceSet[int]] = st.frozensets(
    st.integers(), min_size=1
).map(EquivalenceSet)
st.register_type_strategy(EquivalenceSet, equivalence_sets)

type_of_bounds = st.sampled_from([Included.safer_new, Excluded])

range_constructors = st.sampled_from(
    [
        Range.new_open_right,
        Range.new_open_left,
        Range.new_closed,
        Range.new_open,
    ]
)


def float_ranges(inner=None, outer=None, *, kinds=None):
    """Generate instances of Range[float].

    If `inner` is not None, it must be a Range[float], and all generated
    ranges will be super-sets of `inner`.

    If `outer` is not None, it must be a Range[float], and all generated
    ranges will be sub-sets of `outer`.

    If `kinds` is not None, it must a strategy which constructors of ranges
    for the kind you want.  If None, we produce any kind of range.

    """
    return generalized_ranges(
        inner, outer, RangeType=FloatRangeHelper, kinds=kinds
    )


def date_ranges(inner=None, outer=None, *, kinds=None):
    """Generate instances of Range[datetime].

    If `inner` is not None, it must be a Range[datetime], and all generated
    ranges will be super-sets of `inner`.

    If `outer` is not None, it must be a Range[datetime], and all generated
    ranges will be sub-sets of `outer`.

    If `kinds` is not None, it must a strategy which constructors of ranges
    for the kind you want.  If None, we produce any kind of range.

    """
    return generalized_ranges(
        inner, outer, RangeType=DateRangeHelper, kinds=kinds
    )


def members_of_float_ranges(first_range, *other_ranges):
    """Generates members of any of the ranges passed."""
    ranges = [range_ for range_ in (first_range,) + other_ranges if range_]
    if not ranges:
        raise ValueError("At least a range must be non-empty")
    return reduce(
        operator.or_, (get_float_member_strategy(range_) for range_ in ranges)
    )


def members_of_date_ranges(first_range, *other_ranges):
    """Generates members of any of the ranges passed."""
    ranges = [range_ for range_ in (first_range,) + other_ranges if range_]
    if not ranges:
        raise ValueError("At least a range must be non-empty")
    return reduce(
        operator.or_, (get_date_member_strategy(range_) for range_ in ranges)
    )


def many_float_ranges(n=2, outer=None):
    """Generates `n` ranges in a tuple.

    Ranges are ordered by lowerbound (except empty ranges which have no
    particular ordering key).

    There's no overlapping of generated ranges, nor there's a gap between
    consecutive ones.  No empty ranges are generated.

    If `outer` is not None, all ranges will be contained within this `outer`
    range.  The join of all ranges doesn't necessarily cover the entirety of
    `outer`.

    """
    if outer is None:
        outer = Range.new_open_right(
            FloatRangeHelper.LOWER_LIMIT, FloatRangeHelper.UPPER_LIMIT
        )
    if not outer:
        raise ValueError(f"Cannot generate many ranges within {outer!r}")
    assert n >= 0, f"Invalid value for n: {n}"
    if not n:
        return st.just(())
    # To define n consecutive, non-overlapping ranges we need n + 1 *points*
    # which are members of outer.
    return st.lists(
        get_float_member_strategy(outer, allow_upper=True),
        min_size=n + 1,
        max_size=n + 1,
        unique=lambda x: x,
    ).map(_build_ranges)


def many_date_ranges(n=2, outer=None):
    "Just like many_ranges but generates ranges over dates."
    if outer is None:
        outer = Range.new_open_right(
            DateRangeHelper.LOWER_LIMIT, DateRangeHelper.UPPER_LIMIT
        )
    if not outer:
        raise ValueError(f"Cannot generate many ranges within {outer!r}")
    assert n >= 0, f"Invalid value for n: {n}"
    if not n:
        return st.just(())
    # To define n consecutive, non-overlapping ranges we need n + 1 *points*
    # which are members of outer.
    return st.lists(
        get_date_member_strategy(outer, allow_upper=True),
        min_size=n + 1,
        max_size=n + 1,
        unique=lambda x: x,
    ).map(_build_ranges)


def _build_ranges(points: Sequence[TOrd]) -> Sequence[Range[TOrd]]:
    lower, upper, *points = sorted(points)
    res: List[Range[TOrd]] = [Range.new_open_right(lower, upper)]
    for point in points:
        lower = res[-1].upperbound
        res.append(Range.new_open_right(lower, point))
    return tuple(res)


def float_intervals(min_bases=0, max_bases=None, outer=None):
    """Generates instances of IntervalSet[float].

    If `min_bases` is not 0, we guarantee that the generated intervals will
    have at least min_bases non-empty ranges.  If `min_bases` is 0 the
    intervals may be empty.  If `max_bases` is None, it will default to twice
    the value of `min_bases` (or DEFAULT_MAX_BASES, if `min_bases` is 0).

    If `outer` is not None, the bases of the interval will all be contained
    with `outer`.

    See `many_float_ranges`:func:.

    """
    return generalized_intervals(
        min_bases=min_bases,
        max_bases=max_bases,
        outer=outer,
        many_typed_ranges=many_date_ranges,
    )


def date_intervals(min_bases=0, max_bases=10, outer=None):
    """Generates instances of IntervalSet[datetime].

    The meaning of parameters is the same as in `float_intervals`:func:.

    """
    return generalized_intervals(
        min_bases=min_bases,
        max_bases=max_bases,
        outer=outer,
        many_typed_ranges=many_date_ranges,
    )


def generalized_intervals(
    min_bases=0, max_bases=None, outer=None, *, many_typed_ranges
):
    # The algorithm relies on `many_*_ranges` to generate 2*n - 1 sub-ranges
    # within `outer` and then simply takes for the bases of the intervals
    # those in even-indexes.
    if min_bases is None:
        min_bases = 0
    assert 0 <= min_bases
    if max_bases is None:
        max_bases = DEFAULT_MAX_BASES if not min_bases else 2 * min_bases
    return (
        st.integers(min_value=min_bases or 0, max_value=max_bases)
        .flatmap(
            lambda n: many_typed_ranges(2 * n - 1 if n else 0, outer=outer)
        )
        .map(
            lambda rs: IntervalSet([r for i, r in enumerate(rs) if i % 2 == 0])
        )
    )


DEFAULT_MAX_BASES = 5
st.register_type_strategy(IntervalSet, float_intervals() | date_intervals())


class _RangeHelper:
    lowerbound = DelegatedAttribute("base_range", "lowerbound")
    upperbound = DelegatedAttribute("base_range", "upperbound")

    def __init__(self, lower, upper):
        self.base_range = Range.new_open_right(lower, upper)

    def get_member_strategy(self, allow_upper=False):
        if not self.base_range:
            # Empty ranges have no members, but the generic implementation of
            # Range allows for infinite values of Range which are empty (for all
            # value a, Range(a, a) is empty).
            #
            # Instead of requiring the base_range to be non-empty, we simply
            # generate the lowerbound.
            return st.just(self.base_range.lowerbound)
        if self.lowerbound is -Infinity:
            _min = self.LOWER_LIMIT
        else:
            _min = self.lowerbound
        if self.upperbound is Infinity:
            _max = self.UPPER_LIMIT
        else:
            _max = self.upperbound
        assert not isinstance(_min, type(Infinity)), f"Invalid min {_min}"
        assert not isinstance(_max, type(Infinity)), f"Invalid max {_max}"
        result = self.BASE_STRATEGY(min_value=_min, max_value=_max)
        if isinstance(self.base_range._lower, Excluded):
            result = result.filter(lambda v: v != _min)
        if isinstance(self.base_range._upper, Excluded):
            result = result.filter(lambda v: v != _max)
        if self.lowerbound is -Infinity:
            result = result | st.just(-Infinity)
        if self.upperbound is Infinity and allow_upper:
            result = result | st.just(Infinity)
        return result


class FloatRangeHelper(_RangeHelper):
    """Provides utilities to generate arbitrary ranges within the boundaries of this range."""

    # The lowest number generated unless the range has a lower lowerbound
    LOWER_LIMIT = -1e10

    # The highest number generated unless the range has a higher upperbound
    UPPER_LIMIT = 1e10

    BASE_STRATEGY = staticmethod(
        lambda **kwargs: st.floats(
            allow_infinity=False, allow_nan=False, **kwargs
        )
    )


def get_float_member_strategy(base_range=None, allow_upper=False):
    """Return a strategy to generate members from `base_range`.

    If `base_range` is None, we use a *reduced* base range from -1e10 to 1e10.
    The last number is excluded unless `allow_upper` is True.

    """
    if base_range is None:
        base_range = Range.new_open_right(
            FloatRangeHelper.LOWER_LIMIT, FloatRangeHelper.UPPER_LIMIT
        )
    return FloatRangeHelper(*base_range).get_member_strategy(
        allow_upper=allow_upper
    )


class DateRangeHelper(_RangeHelper):
    """Similar to _Range but generates datetimes."""

    # The lowest datetime generated unless the range has a lower lowerbound
    LOWER_LIMIT = datetime.datetime(1910, 1, 1)

    # The highest datetime generated unless the range has a higher upperbound
    UPPER_LIMIT = datetime.datetime(9900, 12, 31)

    BASE_STRATEGY = staticmethod(
        lambda **kwargs: st.datetimes(
            timezones=st.just(datetime.timezone.utc),
            allow_imaginary=False,
            **kwargs,
        ).map(lambda d: d.replace(tzinfo=None, fold=0))
    )


def get_date_member_strategy(base_range=None, allow_upper=False):
    """Return a strategy to generate members from `base_range`.

    If `base_range` is None, we use a *reduced* base range from 1910-01-01 to
    9900-12-31.  The last date is excluded unless `allow_upper` is True.

    """
    if base_range is None:
        base_range = Range.new_open_right(
            DateRangeHelper.LOWER_LIMIT, DateRangeHelper.UPPER_LIMIT
        )
    return DateRangeHelper(*base_range).get_member_strategy(
        allow_upper=allow_upper
    )


@st.composite
def generalized_ranges(draw, inner=None, outer=None, *, RangeType, kinds=None):
    if kinds is None:
        Kind = draw(range_constructors)
    else:
        Kind = draw(kinds)
    if outer is None:
        # FIXME: Go back to Range(None, None), but since Range(None, None) is
        # an open range, it gets messy to ensure the kind of range I'm
        # generating.
        outer = Kind(RangeType.LOWER_LIMIT, RangeType.UPPER_LIMIT)
    _ensure_range(outer)
    if inner is None:
        midpoint = draw(RangeType(*outer).get_member_strategy())
        inner = Kind(midpoint, midpoint)
    _ensure_range(inner)
    if inner == outer:
        return inner
    if inner > outer:
        raise ValueError(
            f"Impossible to find ranges r such that {inner!r} <= r <= {outer!r}"
        )
    left = RangeType(outer.lowerbound, inner.lowerbound).get_member_strategy()
    if outer.lowerbound is -Infinity:
        left |= st.just(-Infinity)
    right = RangeType(inner.upperbound, outer.upperbound).get_member_strategy()
    if outer.upperbound is Infinity:
        right |= st.just(Infinity)
    # The filter is needed because if right or left are at the point of the
    # boundaries of outer or inner, the arbitrary Kind can create an invalid
    # range.
    #
    # For instance: with outer being `Range(Excluded(0), Excluded(1))`; and
    # Kind being new_closed; we could generate `Range(Included(0),
    # Included(0.5))` which is not contained in outer.
    return draw(
        st.builds(Kind, left, right).filter(lambda r: inner <= r <= outer)
    )


def _ensure_range(which):
    if not isinstance(which, Range):
        raise TypeError(f"Invalid type of '{which!r}', expected a Range")


_ranges_strategies = st.sampled_from([many_float_ranges, many_date_ranges])
st.register_type_strategy(Range, float_ranges() | date_ranges())



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
import doctest
from dataclasses import dataclass

from hypothesis import given
from hypothesis import strategies as s

import xotless.immutables
from xotless.immutables import ImmutableWrapper

from .support import captured_stderr, captured_stdout


def test_doctests():
    with captured_stdout() as stdout, captured_stderr() as stderr:
        failure_count, test_count = doctest.testmod(
            xotless.immutables,
            verbose=True,
            optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS,
            raise_on_error=False,
        )
    if test_count and failure_count:  # pragma: no cover
        print(stdout.getvalue())
        print(stderr.getvalue())
        raise AssertionError("ImmutableWrapper doctest failed")


def test_wrap_callable():
    class Bar:
        def return_self(self):
            return self

    class Foo(Bar):
        pass

    obj = Foo()
    wrapper = ImmutableWrapper(obj)
    assert wrapper.return_self() is obj

    wrapper2 = ImmutableWrapper(obj, wrap_callables=True)
    assert wrapper2.return_self() is wrapper2


def test_wrap_properties():
    from dataclasses import dataclass
    from datetime import datetime, timedelta

    @dataclass(frozen=True)
    class Commodity:
        start_date: datetime
        duration: timedelta

        @property
        def end_date(self):
            return self.start_date + self.duration

    now = datetime.utcnow()
    day = timedelta(1)
    item = ImmutableWrapper(
        Commodity(None, None), wrap_descriptors=True
    ).replace(start_date=now, duration=day)
    assert item.end_date == now + day

    commodity = Commodity(datetime.utcnow(), timedelta(1))
    assert hash(commodity) == hash(ImmutableWrapper(commodity))


def test_wrap_descriptors():
    class Descriptor:
        def __get__(self, instance, owner):
            if instance is None:
                return self
            return id(instance)

    class Item:
        identity = Descriptor()

    item = ImmutableWrapper(Item(), wrap_descriptors=True)
    assert item.identity == id(item)


anything = s.integers() | s.floats() | s.text()
anything = (
    anything
    | s.dictionaries(s.text(), anything)
    | s.tuples(anything)
    | s.lists(anything)
)


@dataclass
class Obj:
    truth: bool

    def __bool__(self):
        return self.truth


@given(anything | s.from_type(Obj))
def test_boolean_values_basic(val):
    assert bool(ImmutableWrapper(val)) == bool(val)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from collections.abc import Mapping
from functools import partial
from typing import Type

import immutables

from .pickablenv import PickableRecordset

# I won't depend on Odoo, but if it installed ImmutableWrapper will be
# intelligent about it.
try:
    from odoo import models

    BaseModel: Type = models.BaseModel
except ImportError:

    class BaseModel:  # type: ignore
        pass


class ImmutableWrapper:
    """Wraps a `target` object so that it can be used immutably.

    Some external objects (most likely Odoo recordsets) are used within the
    pricing programs immutably but *copies* of such objects are created (e.g
    by splitters) while updating or changing some of its attributes.

    Example usage:

    .. code-block:: python

       >>> Commodity = self.env['operation.commodity']                         # doctest: +SKIP
       >>> target = Commodity.search([('start_date', '!=', False)], limit=1)   # doctest: +SKIP

       >>> wrapped = ImmutableWrapper(target)                                  # doctest: +SKIP
       >>> new = wrapped.replace(start_date=target.start_date + timedelta(1),  # doctest: +SKIP
       ...                       duration=timedelta(1))

       >>> assert new.start_date == target.start_date + timedelta(1)           # doctest: +SKIP

    Immutable wrappers are always flattened.  If the `target` object is
    another instance of ImmutableWrapper, we extract its enclosed object and
    "inherit" the overrides:

    .. code-block:: python

       >>> class new:
       ...     pass

       >>> target = new()
       >>> wrapper = ImmutableWrapper(target, overrides=dict(a="a", c="c"))
       >>> wrapper2 = ImmutableWrapper(wrapper, overrides=dict(b="b"))

       >>> wrapper2._ImmutableWrapper__target is target
       True

       >>> wrapper2.a, wrapper2.b, wrapper2.c
       ('a', 'b', 'c')

    Trying to mutate an attribute raises a TypeError:

    .. code-block:: python

       >>> wrapper2.a = 1
       Traceback (most recent call last):
       ...
       TypeError: Cannot mutate an immutable wrapped object...

    If `wrap_callables` is True, then calling method on the target will
    receive the wrapper.

    If `wrap_descriptors` is True, the method ``__get__`` of descriptors in
    the target will receive the wrapper.  This allows properties to work with
    overridden values.

    ImmutableWrapper implements ``__getitem__``.  Any error raised by the
    target is not handled.  The result of __getitem__ might be rewrapped (for
    Odoo recordsets) in another ImmutableWrapper that inherits overrides and
    the value of `wrap_callables` and `wrap_descriptors`.

    ImmutableWrapper implements ``__bool__`` trying to call the underlying's
    type ``__bool__`` with the wrapper as self.  If a TypeError occurrs while
    calling the target's type ``__iter__``, fallback to ``bool(target)``.

    .. versionchanged:: 1.2.0  Add argument `wrap_descriptors`.

    .. versionchanged:: 1.8.0  Add support for ``__getitem__``.

    .. versionchanged:: 2.1.0  Implement ``__bool__``.

    """

    def __new__(
        self,
        target,
        *,
        overrides=None,
        wrap_callables=False,
        wrap_descriptors=False,
    ):
        if overrides is None:
            overrides = {}
        if isinstance(target, PickableRecordset):
            target = target.instance
        if isinstance(target, ImmutableWrapper):
            attrs = dict(target.__overrides, **overrides)
            target = target.__target
        else:
            attrs = overrides
        result = super().__new__(self)
        # Avoid calling __setattr__ which would fail
        result.__dict__["_ImmutableWrapper__target"] = target
        result.__dict__["_ImmutableWrapper__overrides"] = immutables.Map(attrs)
        result.__dict__["_ImmutableWrapper__cache"] = {}
        result.__dict__["_ImmutableWrapper__wrap_callables"] = wrap_callables
        result.__dict__[
            "_ImmutableWrapper__wrap_descriptors"
        ] = wrap_descriptors
        return result

    def __getitem__(self, key):
        item = self.__target[key]  # let errors go
        return self.__maybe_rewrap(self, item, keep_overrides=True)

    def __getattr__(self, attr):
        from xotl.tools.symbols import Unset

        res = self.__overrides.get(attr, Unset)
        if res is Unset:
            res = self.__cache.get(attr, Unset)
        if res is Unset and self.__wrap_descriptors:
            descriptor = getattr(type(self.__target), attr, None)
            if descriptor is not None:
                fget = getattr(descriptor, "__get__", None)
                if fget is not None:
                    try:
                        res = fget(self, type(self.__target))
                    except TypeError:
                        res = Unset
        if res is Unset:
            res = getattr(self.__target, attr)
        res = self.__maybe_rewrap(self, res, keep_overrides=False)
        self.__cache[attr] = res
        return res

    def __bool__(self):
        if type(self.__target) not in (int, float):
            # int and float will raise a TypeError, so avoid the call.
            method = getattr(type(self.__target), "__bool__", None)
        else:
            method = None
        if method:
            try:
                return method(self)
            except TypeError:
                pass
        return bool(self.__target)

    def __setattr__(self, attr, value):
        raise TypeError(
            "Cannot mutate an immutable wrapped object.  Use method 'replace'."
        )

    def replace(self, **attrs):
        """Return a new wrapper of the same object with updates."""
        overrides = dict(self.__overrides, **attrs)
        return ImmutableWrapper(
            self.__target,
            overrides=overrides,
            wrap_callables=self.__wrap_callables,
            wrap_descriptors=self.__wrap_descriptors,
        )

    @classmethod
    def __maybe_rewrap(cls, self, res, keep_overrides):
        if (
            self.__wrap_callables
            and callable(res)
            and getattr(res, "__self__", None) is self.__target
        ):
            # If `res` a bound method of __target, extract the underlying
            # fuction so that `self` in the method is the ImmutableWrapper.
            res = partial(res.__func__, self)
        if isinstance(res, BaseModel):
            if not keep_overrides:
                res = ImmutableWrapper(
                    res,
                    wrap_callables=self.__wrap_callables,
                    wrap_descriptors=self.__wrap_descriptors,
                )
            else:
                res = ImmutableWrapper(
                    res,
                    overrides=self.__overrides,
                    wrap_callables=self.__wrap_callables,
                    wrap_descriptors=self.__wrap_descriptors,
                )
        return res

    # IMPORTANT: The __cache can't be part of the hash nor of eq because is
    # just an optimization and not a part of the *identity* of the object.
    #
    # Also, there's a minor problem in __eq__ which allows to compare an
    # ImmutableWrapper with the underlying object; this means that instances
    # of ImmutableWrapper and the target object will hash differently despite
    # being equal.  This is only an issue if you mix ImmutableWrappers with
    # wrapped objects in the same collection (set, dict, etc...)
    def __hash__(self):
        if self.__overrides:
            return hash((self.__target, self.__overrides))
        else:
            return hash(self.__target)

    def __eq__(self, other):
        if isinstance(other, ImmutableWrapper):
            return (
                self.__target == other.__target
                and self.__overrides == other.__overrides
            )
        elif isinstance(other, type(self.__target)):
            # We allow for a wrapper to be equal to its underlying target
            # object, but only if there are no overrides.
            return not self.__overrides and self.__target == other
        else:
            return NotImplemented

    def __repr__(self):
        return (
            f"<ImmutableWrapper of '{self.__target!r}' with {self.__overrides}; "
            f"cache: {self.__cache}; wrap_callables: {self.__wrap_callables}>"
        )

    def __getnewargs__(self):
        if isinstance(self.__target, BaseModel):
            return (PickableRecordset.from_recordset(self.__target),)
        else:
            return (self.__target,)

    def __getstate__(self):
        cache = self.__cache
        overrides = self.__overrides
        return (overrides, cache)

    def __setstate__(self, state):
        overrides, cache = state
        self.__dict__["_ImmutableWrapper__overrides"] = overrides
        self.__dict__["_ImmutableWrapper__cache"] = cache


class ImmutableChainMap(Mapping):  # noqa
    def __init__(self, *maps):
        self.maps = maps
        self._cache = None

    def _fill_cache(self):
        if self._cache is None:
            self._cache = cache = {}
            for mapping in reversed(self.maps):
                cache.update(mapping)

    def __hash__(self):
        return hash(tuple(hash(mapping) for mapping in self.maps))

    def __eq__(self, other):
        if isinstance(other, ImmutableChainMap):
            other._fill_cache()
            self._fill_cache()
            return self._cache == other._cache
        else:
            return NotImplemented

    def __getitem__(self, key):
        self._fill_cache()
        return self._cache[key]

    def __len__(self):
        self._fill_cache()
        return len(self._cache)

    def __iter__(self):
        self._fill_cache()
        return iter(self._cache)

    def __getstate__(self):
        return self.maps

    def __setstate__(self, state):
        self.__init__(*state)
